import socket
import select
from collections import defaultdict
from functools import reduce

from sauna.consumers.base import AsyncConsumer
from sauna import check_results_lock, check_results
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('TCPServer')


@my_consumer.consumer()
class TCPServerConsumer(AsyncConsumer):

    service_checks = {}

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'port': config.get('port', 5555),
            'backlog': config.get('port', 128),
            'keepalive': config.get('keepalive', True)
        }
        self.read_wanted, self.write_wanted = ([], [])
        self.write_buffers = defaultdict(bytes)

    def _get_current_status(self):
        from sauna.plugins.base import Plugin

        def reduce_status(accumulated, update_value):
            if update_value.status > Plugin.STATUS_CRIT:
                return accumulated
            return accumulated if accumulated > update_value.status else \
                update_value.status

        with check_results_lock:
            res = reduce(reduce_status, check_results.values(), 0)

        if res == Plugin.STATUS_OK:
            status = b'OK\n'
        elif res == Plugin.STATUS_WARN:
            status = b'WARNING\n'
        elif res == Plugin.STATUS_CRIT:
            status = b'CRITICAL\n'
        else:
            status = b'UNKNOWN\n'
        return status

    def _create_server(self):
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setblocking(0)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind(('', self.config['port']))
        self.server.listen(self.config['backlog'])

    def _accept_new_connection(self):
        client_socket, address = self.server.accept()
        self.logging(
            'debug', 'New connection from {}'.format(address[0])
        )
        self.write_wanted.append(client_socket)
        self.read_wanted.append(client_socket)
        return client_socket

    def _activate_keepalive(self, s, after_idle_sec=30, interval_sec=10,
                            max_fails=5):
        """Set TCP keepalive on an open socket.

        It activates after 30 second (after_idle_sec) of idleness,
        then sends a keepalive ping once every 10 seconds (interval_sec),
        and closes the connection after 5 failed ping (max_fails).
        """
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)

    def _close_socket(self, s):
        try:
            s.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        try:
            s.close()
        except socket.error:
            pass
        self._remove_from_list(self.write_wanted, s)
        self._remove_from_list(self.read_wanted, s)
        try:
            del self.write_buffers[s]
        except KeyError:
            pass
        self.logging('debug', 'Closed connection')

    @staticmethod
    def _remove_from_list(list_, value):
        while value in list_:
            try:
                list_.remove(value)
            except ValueError:
                pass

    def _handle_read_event(self, s):
        if s == self.server:
            client_socket = self._accept_new_connection()
            if self.config['keepalive']:
                self._activate_keepalive(client_socket)
            self.write_buffers[client_socket] += self._get_current_status()
            self.write_wanted.append(client_socket)
        else:
            try:
                read_data = s.recv(4096)
            except socket.error as e:
                self.logging('warning',
                             'Error while receiving, closing connection: {}'
                             .format(e))
                self._close_socket(s)
                return
            if len(read_data) == 0:
                self._close_socket(s)
            else:
                self.logging('debug', 'Received data')
                if b'\n' in read_data:
                    self.write_buffers[s] += self._get_current_status()
                    self.write_wanted.append(s)

    def _handle_write_event(self, s):
        try:
            sent_len = s.send(self.write_buffers[s])
        except socket.error as e:
            self.logging('warning',
                         'Error while sending, closing connection: {}'
                         .format(e))
            self._close_socket(s)
            return
        self.write_buffers[s] = self.write_buffers[s][sent_len:]
        if not self.write_buffers[s]:
            self.write_wanted.remove(s)
            self.logging('debug', 'Sent data')

    def run(self, must_stop, *args):
        self._create_server()
        self.read_wanted = [self.server]
        self.write_wanted = []

        while not must_stop.is_set():
            readable, writable, errored = select.select(
                self.read_wanted,
                self.write_wanted,
                self.read_wanted + self.write_wanted,
                1
            )

            for s in errored:
                self.logging('warning', 'Connection in error, closing it')
                self._close_socket(s)

            for s in readable:
                self._handle_read_event(s)

            for s in writable:
                self._handle_write_event(s)

        self.logging('debug', 'Exited consumer thread')

    @staticmethod
    def config_sample():
        return '''
        # Listen on a TCP port and serve results to incoming connections
        TCPServer:
          port: 5555
        '''



import socket
import struct
import binascii
from copy import deepcopy
import itertools

from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('NSCA')


def encrypt_xor(data, iv, key):
    for i in (iv, key):
        i = itertools.cycle(i)
        data = bytes(x ^ y for x, y in zip(data, i))
    return data


@my_consumer.consumer()
class NSCAConsumer(QueuedConsumer):

    protocol_version = 3
    max_hostname_size = 64
    max_service_size = 128
    max_output_size = 4096

    init_payload_fmt = '!128sL'
    init_payload_size = struct.calcsize(init_payload_fmt)
    service_payload_fmt = '!hhIIh{}s{}s{}s'.format(
        max_hostname_size, max_service_size, max_output_size
    )
    service_payload_size = struct.calcsize(service_payload_fmt)

    encryption_functions = {
        0: lambda x, y, z: x,
        1: encrypt_xor
    }

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'server': config.get('server', 'localhost'),
            'port': config.get('port', 5667),
            'timeout': config.get('timeout', 10),
            'encryption': config.get('encryption', 0),
            'key': config.get('key', '').encode('ascii'),
        }
        self._last_good_receiver_address = None

    def _recv_init_payload(self, s):
        init_payload = bytes()
        while len(init_payload) < self.init_payload_size:
            buffer = s.recv(self.init_payload_size - len(init_payload))
            if buffer:
                init_payload += buffer
        return self._decode_init_payload(init_payload)

    def _decode_init_payload(self, init_payload):
        return struct.unpack(self.init_payload_fmt, init_payload)

    def _encode_service_payload(self, service_check):
        service_payload_list = [
            self.protocol_version,
            0,  # Padding
            0,  # Placeholder for CRC
            service_check.timestamp,
            service_check.status,
            service_check.hostname.encode('utf8'),
            service_check.name.encode('utf8'),
            service_check.output.encode('utf8')
        ]
        crc = binascii.crc32(struct.pack(self.service_payload_fmt,
                                         *service_payload_list))
        service_payload_list[2] = crc
        return struct.pack(self.service_payload_fmt, *service_payload_list)

    def _format_service_check(self, service_check):
        # C strings are null terminated, we need one extra char for each
        if len(service_check.hostname) > self.max_hostname_size - 1:
            raise ValueError('NSCA hostnames can be up to {} characters'.
                             format(self.max_hostname_size - 1))
        if len(service_check.name) > self.max_service_size - 1:
            raise ValueError('NSCA service names can be up to {} characters'.
                             format(self.max_service_size - 1))
        # Silently truncate output to its max length
        if len(service_check.output) > self.max_output_size - 1:
            service_check = deepcopy(service_check)
            truncate_output = self.max_output_size - 1
            service_check.output = service_check.output[:truncate_output]
        return service_check

    def _encrypt_service_payload(self, service_payload, iv):
        try:
            encryption_mode = self.config['encryption']
            encryption_function = self.encryption_functions[encryption_mode]
        except KeyError:
            raise ValueError('Encryption mode not supported')
        data = encryption_function(service_payload, iv, self.config['key'])
        return data

    def _get_receivers_addresses(self):
        """Retrieve all the addresses associated with a hostname.

        It will return in priority the address of the last known receiver which
        accepted the previous check.
        """
        receivers = socket.getaddrinfo(
            self.config['server'], self.config['port'],
            proto=socket.IPPROTO_TCP
        )
        # Only keep the actual address
        addresses = [r[4][0] for r in receivers]
        try:
            addresses.remove(self._last_good_receiver_address)
            addresses = [self._last_good_receiver_address] + addresses
        except ValueError:
            pass
        return addresses

    def _send_to_receiver(self, service_check, receiver_address):
        with socket.socket() as s:
            s.settimeout(self.config['timeout'])
            s.connect((receiver_address, self.config['port']))
            iv, timestamp = self._recv_init_payload(s)
            service_payload = self._encode_service_payload(service_check)
            s.sendall(self._encrypt_service_payload(service_payload, iv))

    def _send(self, service_check):
        for receiver_address in self._get_receivers_addresses():
            try:
                self._send_to_receiver(service_check, receiver_address)
                self._last_good_receiver_address = receiver_address
                return
            except OSError as e:
                self.logging('info', 'Could not send check to receiver {}: '
                             '{}'.format(receiver_address, e))
        raise IOError('No receiver accepted the check')

    @staticmethod
    def config_sample():
        return '''
        # Send service check to a NSCA server
        # Only encryption methods 0 and 1 are supported
        # Max plugin output is 4096 bytes
        NSCA:
          server: receiver.shinken.tld
          port: 5667
          timeout: 10
          encryption: 1
          key: verylongkey
        '''



from sauna.plugins.base import Plugin


def bytes_to_human(n):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return '%.1f%s' % (value, s)
    return "%sB" % n


def human_to_bytes(size):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    size = size.upper()
    for i, s in enumerate(symbols):
        if s in size:
            base = int(size.split(s)[0])
            return base * (1024 ** (i+1))
    return int(size)


class PluginRegister:
    all_plugins = {}

    def __init__(self, name):
        self.name = name
        self.plugin_class = None
        self.checks = {}

    def check(self, **options):
        def decorator(func):
            check_name = options.pop("name", func.__name__)
            self.checks[check_name] = func
            return func
        return decorator

    def plugin(self):
        def decorator(plugin_cls):
            self.plugin_class = plugin_cls
            # At this point all data should be set
            self.all_plugins[self.name] = {
                'plugin_cls': self.plugin_class,
                'checks': self.checks
            }
            return plugin_cls
        return decorator

    @classmethod
    def get_plugin(cls, name):
        try:
            return cls.all_plugins[name]
        except KeyError:
            return None



from collections import namedtuple
import threading
import queue
import logging
import time
import socket
import os
import textwrap
import signal
import importlib
import pkgutil


from sauna import plugins, consumers
from sauna.plugins.base import Check
from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister
from sauna.plugins import PluginRegister

__version__ = '0.0.4'

ServiceCheck = namedtuple('ServiceCheck',
                          ['timestamp', 'hostname', 'name',
                           'status', 'output'])

# Global dict containing the last status of each check
# Needs to hold a lock to access it
check_results = {}
check_results_lock = threading.Lock()

try:
    # In Python 3.2 threading.Event is a factory function
    # the real class lives in threading._Event
    event_type = threading._Event
except AttributeError:
    event_type = threading.Event


class DependencyError(Exception):

    def __init__(self, plugin, dep_name, pypi='', deb=''):
        self.msg = '{} depends on {}. It can be installed with:\n'.format(
            plugin, dep_name
        )
        if pypi:
            self.msg = '{}    pip install {}\n'.format(self.msg, pypi)
        if deb:
            self.msg = '{}    apt-get install {}\n'.format(self.msg, deb)

    def __str__(self):
        return self.msg


def read_config(config_file):
    # importing yaml here because dependency is not installed
    # when fetching __version__ from setup.py
    import yaml

    try:
        with open(config_file) as f:
            return yaml.safe_load(f)
    except OSError as e:
        print('Cannot read configuration file {}: {}'
              .format(config_file, e))
        exit(1)


class Sauna:

    def __init__(self, config=None):
        if config is None:
            config = {}
        self.config = config

        self.must_stop = threading.Event()
        self._consumers_queues = []

    @classmethod
    def assemble_config_sample(cls, path):
        sample = '---\nperiodicity: 120\nhostname: node-1.domain.tld\n'

        sample += '\nconsumers:\n'
        consumers_sample = ''
        for _, consumer_info in ConsumerRegister.all_consumers.items():
            if hasattr(consumer_info['consumer_cls'], 'config_sample'):
                consumers_sample += textwrap.dedent(
                    consumer_info['consumer_cls'].config_sample()
                )
        sample += consumers_sample.replace('\n', '\n  ')

        sample += '\nplugins:\n'
        plugins_sample = ''
        for _, plugin_info in PluginRegister.all_plugins.items():
            if hasattr(plugin_info['plugin_cls'], 'config_sample'):
                plugins_sample += textwrap.dedent(
                    plugin_info['plugin_cls'].config_sample()
                )
        sample += plugins_sample.replace('\n', '\n  ')

        file_path = os.path.join(path, 'sauna-sample.yml')
        with open(file_path, 'w') as f:
            f.write(sample)

        return file_path

    @property
    def hostname(self):
        return self.config.get('hostname', socket.getfqdn())

    @property
    def periodicity(self):
        return self.config.get('periodicity', 120)

    def get_active_checks_name(self):
        checks = self.get_all_active_checks()
        return [check.name for check in checks]

    def get_all_available_consumers(self):
        return_consumers = []
        for plugin_name, _ in ConsumerRegister.all_consumers.items():
            return_consumers.append(plugin_name)

        return return_consumers

    def get_all_available_checks(self):
        checks = {}
        for plugin_name, data in PluginRegister.all_plugins.items():
            checks[plugin_name] = []
            for check in data['checks']:
                checks[plugin_name].append(check)
        return checks

    def get_all_active_checks(self):
        checks = []
        deps_error = []
        for plugin_name, plugin_data in self.config['plugins'].items():

            # Load plugin
            plugin_info = PluginRegister.get_plugin(plugin_name)
            if not plugin_info:
                print('Plugin {} does not exist'.format(plugin_name))
                exit(1)

            # Configure plugin
            try:
                plugin = plugin_info['plugin_cls'](
                    plugin_data.get('config', {})
                )
            except DependencyError as e:
                deps_error.append(str(e))
                continue

            # Launch plugin checks
            for check in plugin_data['checks']:
                check_func = getattr(plugin, check['type'])

                if not check_func:
                    print('Unknown check {} on plugin {}'.format(check['type'],
                                                                 plugin_name))
                    exit(1)

                check_name = (check.get('name') or '{}_{}'.format(
                    plugin_name, check['type']
                ).lower())

                checks.append(Check(check_name, check_func, check))
        if deps_error:
            for error in deps_error:
                print(error)
            exit(1)
        return checks

    def launch_all_checks(self, hostname):
        for check in self.get_all_active_checks():

            try:
                status, output = check.run_check()
            except Exception as e:
                logging.warning('Could not run check {}: {}'.format(
                    check.name, str(e)
                ))
                status = 3
                output = str(e)
            s = ServiceCheck(
                timestamp=int(time.time()),
                hostname=hostname,
                name=check.name,
                status=status,
                output=output
            )
            yield s

    def run_producer(self):
        while True:
            for service_check in self.launch_all_checks(self.hostname):
                logging.debug(
                    'Pushing to main queue: {}'.format(service_check))
                self.send_data_to_consumers(service_check)
                with check_results_lock:
                    check_results[service_check.name] = service_check
            if self.must_stop.wait(timeout=self.periodicity):
                break
        logging.debug('Exited producer thread')

    def term_handler(self, *args):
        """Notify producer and consumer that they should stop."""
        if not self.must_stop.is_set():
            self.must_stop.set()
            self.send_data_to_consumers(self.must_stop)
            logging.info('Exiting...')

    def send_data_to_consumers(self, data):
        for queue in self._consumers_queues:
            queue.put(data)

    def launch(self):
        # Start producer and consumer threads
        producer = threading.Thread(
            name='producer', target=self.run_producer
        )
        producer.start()

        consumers_threads = []
        for consumer_name, consumer_config in self.config['consumers'].items():

            consumer_info = ConsumerRegister.get_consumer(consumer_name)
            if not consumer_info:
                print('Plugin {} does not exist'.format(consumer_name))
                exit(1)

            try:
                consumer = consumer_info['consumer_cls'](consumer_config)
            except DependencyError as e:
                print(str(e))
                exit(1)

            if isinstance(consumer, QueuedConsumer):
                consumer_queue = queue.Queue()
                self._consumers_queues.append(consumer_queue)
            else:
                consumer_queue = None

            consumer_thread = threading.Thread(
                name='consumer_{}'.format(consumer_name),
                target=consumer.run,
                args=(self.must_stop, consumer_queue)
            )

            consumer_thread.start()
            consumers_threads.append(consumer_thread)
            logging.debug(
                'Running {} with {}'.format(consumer_name, consumer_config)
            )

        signal.signal(signal.SIGTERM, self.term_handler)
        signal.signal(signal.SIGINT, self.term_handler)

        producer.join()
        self.term_handler()

        for consumer_thread in consumers_threads:
            consumer_thread.join()

        logging.debug('Exited main thread')


def get_all_subclass(main_class):
    # Check all subclass with recursion
    subclasses = []
    for subclass in main_class.__subclasses__():
        subclasses.append(subclass)
        subclasses += get_all_subclass(subclass)
    return tuple(subclasses)


def import_submodules(package):
    # Please don't use logger in this method,
    # it will disable all logger parameters
    if isinstance(package, str):
        package = importlib.import_module(package)

    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        if not name.startswith("_") and is_pkg is False:
            full_name = package.__name__ + '.' + name
            importlib.import_module(full_name)

import_submodules(__name__+'.plugins.ext')
import_submodules(__name__+'.consumers.ext')



from sauna.plugins.base import PsutilPlugin
from sauna.plugins import human_to_bytes, bytes_to_human, PluginRegister

my_plugin = PluginRegister('Memory')


@my_plugin.plugin()
class Memory(PsutilPlugin):

    def __init__(self, config):
        super().__init__(config)
        self._virtual_memory = None
        self._swap_memory = None

    @my_plugin.check()
    def available(self, check_config):
        available = self.virtual_memory.available
        return (
            self._value_to_status_more(available, check_config,
                                       human_to_bytes),
            'Memory available: {}'.format(bytes_to_human(available))
        )

    @my_plugin.check()
    def used_percent(self, check_config):
        used_percent = self.virtual_memory.percent
        check_config = self._strip_percent_sign_from_check_config(check_config)
        return (
            self._value_to_status_less(used_percent, check_config),
            'Memory used: {}%'.format(used_percent)
        )

    @my_plugin.check()
    def swap_used_percent(self, check_config):
        swap_used_percent = self.swap_memory.percent
        check_config = self._strip_percent_sign_from_check_config(check_config)
        return (
            self._value_to_status_less(swap_used_percent, check_config),
            'Swap used: {}%'.format(swap_used_percent)
        )

    @property
    def virtual_memory(self):
        if not self._virtual_memory:
            self._virtual_memory = self.psutil.virtual_memory()
        return self._virtual_memory

    @property
    def swap_memory(self):
        if not self._swap_memory:
            self._swap_memory = self.psutil.swap_memory()
        return self._swap_memory

    @staticmethod
    def config_sample():
        return '''
        # System memory
        Memory:
          checks:
            - type: available
              warn: 6G
              crit: 2G
            - type: used_percent
              warn: 80%
              crit: 90%
            - type: swap_used_percent
              warn: 50%
              crit: 70%
        '''



from sauna.plugins import PluginRegister
from sauna.plugins.base import PsutilPlugin

my_plugin = PluginRegister('Processes')


@my_plugin.plugin()
class Processes(PsutilPlugin):

    @my_plugin.check()
    def count(self, check_config):
        num_pids = len(self.psutil.pids())
        return (
            self._value_to_status_less(num_pids, check_config),
            '{} processes'.format(num_pids)
        )

    @my_plugin.check()
    def running(self, check_config):
        process_exec = check_config['exec']
        required_args = check_config.get('args', '').split()

        for process in self.psutil.process_iter():
            try:
                if process.exe() != process_exec:
                    continue
                if self._required_args_are_in_cmdline(required_args,
                                                      process.cmdline()):
                    return (
                        self.STATUS_OK,
                        'Process {} is running'.format(process_exec)
                    )
            except (self.psutil.NoSuchProcess, self.psutil.AccessDenied):
                # Zombies and processes that stopped throw NoSuchProcess
                pass
        return (
            self.STATUS_CRIT,
            'Process {} is not running'.format(process_exec)
        )

    def _required_args_are_in_cmdline(self, required_args, cmdline):
        for arg in required_args:
            if arg not in cmdline[1:]:
                return False
        return True

    @staticmethod
    def config_sample():
        return '''
        # Information about processes
        Processes:
          checks:
            # Number of processes in the system
            - type: count
              warn: 400
              crit: 500
            # Critical if process is not running
            - type: running
              name: docker_running
              exec: /usr/bin/docker
              args: daemon
        '''



from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('HTTP')


@my_consumer.consumer()
class HTTPConsumer(QueuedConsumer):

    def __init__(self, config):
        super().__init__(config)
        try:
            import requests
            self.requests = requests
        except ImportError:
            from .. import DependencyError
            raise DependencyError(self.__class__.__name__, 'requests',
                                  'requests', 'python3-requests')
        self.config = {
            'url': config.get('url', 'http://localhost'),
            'timeout': config.get('timeout', 60),
            'headers': config.get('headers', None)
        }

    def _send(self, service_check):
        data = {
            'timestamp': service_check.timestamp,
            'hostname': service_check.hostname,
            'service': service_check.name,
            'status': service_check.status,
            'output': service_check.output
        }
        response = self.requests.post(
            self.config['url'], timeout=self.config['timeout'],
            headers=self.config['headers'], json=data
        )
        response.raise_for_status()

    @staticmethod
    def config_sample():
        return '''
        # Posts a service check trough HTTP
        # Payload is serialized in JSON
        HTTP:
          url: http://server.tld/services
          timeout: 60
          headers:
            X-Auth-Token: XaiZevii0thaemaezaeJ
        '''



from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('Stdout')


@my_consumer.consumer()
class StdoutConsumer(QueuedConsumer):

    def _send(self, service_check):
        print(service_check)

    @staticmethod
    def config_sample():
        return '''
        # Just prints checks on the standard output
        Stdout:
        '''



import time
import logging


class Consumer:

    def __init__(self, config):
        if config is None:
            config = {}
        self.stale_age = config.get('stale_age', 300)
        self.retry_delay = 10

    def logging(self, lvl, message):
        log = getattr(logging, lvl)
        message = '[{}] {}'.format(self.__class__.__name__, message)
        log(message)

    def run(self, must_stop, queue):
        """Method to override in consumers."""
        raise NotImplemented()


class QueuedConsumer(Consumer):
    """Consumer that processes checks synchronously.

    QueuedConsumers wait for checks to appear on a queue. They process each
    check in order until the queue is empty.
    """
    def _send(self, service_check):
        """Method to override in consumers."""
        raise NotImplemented()

    def try_send(self, service_check, must_stop):

        while True:
            if must_stop.is_set():
                return
            if service_check.timestamp + self.stale_age < int(time.time()):
                self.logging('warning', 'Dropping check because it is too old')
                return
            try:
                self._send(service_check)
                self.logging('info', 'Check sent')
                return
            except Exception as e:
                self.logging('warning', 'Could not send check: {}'.format(e))
                if must_stop.is_set():
                    return
                self._wait_before_retry(must_stop)

    def _wait_before_retry(self, must_stop):
        self.logging('info',
                     'Waiting {}s before retry'.format(self.retry_delay))
        for i in range(self.retry_delay):
            if must_stop.is_set():
                return
            time.sleep(1)

    def run(self, must_stop, queue):
        from sauna import event_type
        while not must_stop.is_set():
            service_check = queue.get()
            if isinstance(service_check, event_type):
                continue
            self.logging(
                'debug',
                'Got check {}'.format(service_check)
            )
            self.try_send(service_check, must_stop)
        self.logging('debug', 'Exited consumer thread')


class AsyncConsumer(Consumer):
    """Consumer that processes checks asynchronously.

    It is up to the consumer to read the checks when it needs to. No
    queueing is made.
    """
    pass



import copy


class Plugin:
    """Base class to implement check plugins.

    Most methods use a check_config dict that mostly defines thresholds.
    It looks like:
    check_config = {
        'warn': '80%',
        'crit': '90%',
        'other_data': 'something'
    }
    """

    STATUS_OK = 0
    STATUS_WARN = 1
    STATUS_CRIT = 2
    STATUS_UNKNOWN = 3

    def __init__(self, config):
        if config is None:
            config = {}
        self.config = config

    @classmethod
    def get_thresholds(cls, check_config, modifier=None):
        critical_threshold = check_config['crit']
        if modifier:
            critical_threshold = modifier(critical_threshold)
        warning_threshold = check_config['warn']
        if modifier:
            warning_threshold = modifier(warning_threshold)
        return critical_threshold, warning_threshold

    @classmethod
    def _value_to_status_less(cls, value, check_config, modifier=None):
        """Return an error when the value should be less than threshold."""
        critical, warning = cls.get_thresholds(check_config, modifier)
        if value >= critical:
            return cls.STATUS_CRIT
        elif value >= warning:
            return cls.STATUS_WARN
        else:
            return cls.STATUS_OK

    @classmethod
    def _value_to_status_more(cls, value, check_config, modifier=None):
        """Return an error when the value should be more than threshold."""
        critical, warning = cls.get_thresholds(check_config, modifier)
        if value <= critical:
            return cls.STATUS_CRIT
        elif value <= warning:
            return cls.STATUS_WARN
        else:
            return cls.STATUS_OK

    @classmethod
    def _strip_percent_sign(cls, value):
        try:
            return int(value)
        except ValueError:
            return int(value.split('%')[0])

    @classmethod
    def _strip_percent_sign_from_check_config(cls, check_config):
        check_config = copy.deepcopy(check_config)
        check_config['warn'] = cls._strip_percent_sign(check_config['warn'])
        check_config['crit'] = cls._strip_percent_sign(check_config['crit'])
        return check_config


class Check:
    def __init__(self, name, check_func, config):
        self.name = name
        self.check_func = check_func
        self.config = config

    def run_check(self):
        return self.check_func(self.config)


class PsutilPlugin(Plugin):

    def __init__(self, config):
        super().__init__(config)
        try:
            import psutil
            self.psutil = psutil
        except ImportError:
            from .. import DependencyError
            raise DependencyError(self.__class__.__name__, 'psutil',
                                  'psutil', 'python3-psutil')



class ConsumerRegister:
    all_consumers = {}

    def __init__(self, name):
        self.name = name
        self.consumer_class = None

    def consumer(self):
        def decorator(plugin_cls):
            self.consumer_class = plugin_cls
            self.all_consumers[self.name] = {
                'consumer_cls': self.consumer_class
            }
            return plugin_cls
        return decorator

    @classmethod
    def get_consumer(cls, name):
        try:
            return cls.all_consumers[name]
        except KeyError:
            return None



#!/usr/bin/env python3
"""Health checker

Usage:
  sauna [--level=<lvl>] [--config=FILE]
  sauna sample
  sauna list-active-checks [--config=FILE]
  sauna list-available-checks [--config=FILE]
  sauna list-available-consumers [--config=FILE]
  sauna (-h | --help)
  sauna --version

Options:
  -h --help      Show this screen.
  --version      Show version.
  --level=<lvl>  Log level [default: warn].
  --config=FILE  Config file [default: sauna.yml].

"""
import logging

from docopt import docopt

import sauna


def main():
    args = docopt(__doc__, version=sauna.__version__)
    conf_file = args['--config']
    logging.basicConfig(
        format='%(asctime)s - %(levelname)-8s - %(message)s',
        datefmt='%Y/%m/%d %H:%M:%S',
        level=getattr(logging, args['--level'].upper(), 'WARN')
    )
    logging.getLogger('requests').setLevel(logging.ERROR)

    if args.get('sample'):
        sauna_instance = sauna.Sauna()
        file_path = sauna_instance.assemble_config_sample('./')
        print('Created file {}'.format(file_path))
    else:
        config = sauna.read_config(conf_file)
        sauna_instance = sauna.Sauna(config)
        if args.get('list-active-checks'):
            for name in sauna_instance.get_active_checks_name():
                print('{}'.format(name))
        elif args.get('list-available-checks'):
            for plugin, checks in\
                    sauna_instance.get_all_available_checks().items():
                print('{}: {}'.format(plugin, ', '.join(checks)))
        elif args.get('list-available-consumers'):
            for consumer in sauna_instance.get_all_available_consumers():
                print('{}'.format(consumer))
        else:
            sauna_instance.launch()

    logging.shutdown()



import time
from datetime import timedelta

from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('PuppetAgent')


@my_plugin.plugin()
class PuppetAgent(Plugin):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'summary_path': config.get(
                'summary_path',
                '/var/lib/puppet/state/last_run_summary.yaml'
            )
        }
        self._last_run_summary = None

    @property
    def last_run_summary(self):
        import yaml
        if not self._last_run_summary:
            with open(self.config['summary_path']) as f:
                self._last_run_summary = yaml.safe_load(f)
        return self._last_run_summary

    @my_plugin.check()
    def last_run_delta(self, check_config):
        current_time = int(time.time())
        last_run_time = self.last_run_summary['time']['last_run']
        delta = current_time - last_run_time
        status = self._value_to_status_less(delta, check_config)
        output = 'Puppet last ran {} ago'.format(timedelta(seconds=delta))
        return status, output

    @my_plugin.check()
    def failures(self, check_config):
        failures = self.last_run_summary['events']['failure']
        status = self._value_to_status_less(failures, check_config)
        if failures:
            output = 'Puppet last run had {} failure(s)'.format(failures)
        else:
            output = 'Puppet ran without trouble'
        return status, output

    @staticmethod
    def config_sample():
        return '''
        # Puppet agent
        # sauna user must be able to read:
        # /var/lib/puppet/state/last_run_summary.yaml
        PuppetAgent:
          checks:
            - type: last_run_delta
              warn: 3600
              crit: 14400
            - type: failures
              warn: 1
              crit: 1
        '''



import socket
import re
import subprocess

from sauna.plugins import Plugin
from sauna.plugins import PluginRegister

my_plugin = PluginRegister('Postfix')


@my_plugin.plugin()
class Postfix(Plugin):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'host': config.get('host', 'localhost'),
            'port': config.get('port', 4280),
            'timeout': config.get('timeout', 5),
            'method': config.get('method', 'mailq')
        }
        self._mailq_output = None

    @my_plugin.check()
    def queue_size(self, check_config):
        queue_size = self._get_queue_size()
        return (self._value_to_status_less(queue_size, check_config),
                '{} mail(s) in queue'.format(queue_size))

    @property
    def mailq_output(self):
        if not self._mailq_output:
            if self.config['method'] == 'tcp':
                self._mailq_output = self._fetch_showq()
            else:
                self._mailq_output = self._exec_mailq_command()
        return self._mailq_output

    def _get_queue_size(self):
        if 'Mail queue is empty' in self.mailq_output:
            return 0
        if 'mail system is down' in self.mailq_output:
            raise Exception('Cannot get queue size: {}'.
                            format(self.mailq_output))
        match = re.search(
            r'^-- \d+ [GMK]bytes in (\d+) Requests?\.$',
            self.mailq_output,
            re.MULTILINE
        )
        if not match:
            raise Exception('Cannot parse mailq')
        return int(match.group(1))

    def _fetch_showq(self):
        """Connect to Postfix showq inet daemon and retrieve queue.

        This method is faster than executing mailq because it doesn't fork
        processes.
        It requires to have showq inet daemon activated which is not the case
        by default. To make showq listen on the loopback interface on port
        4280, add to your master.cf:
        127.0.0.1:4280     inet  n       -       -       -       -       showq
        """
        showq = bytes()
        with socket.create_connection((self.config['host'],
                                       self.config['port']),
                                      timeout=self.config['timeout']) as s:
            while True:
                buffer = bytearray(4096)
                bytes_received = s.recv_into(buffer)
                if bytes_received == 0:
                    break
                showq += buffer
        return showq.decode(encoding='utf-8')

    def _exec_mailq_command(self):
        """Execute mailq command to communicate with showq daemon.

        mailq invokes postqueue with a setuid bit that grant it access to the
        Unix socket of showq daemon.
        """
        p = subprocess.Popen(
            'mailq',
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True
        )
        stdout, _ = p.communicate(timeout=self.config['timeout'])
        return stdout

    @staticmethod
    def config_sample():
        return '''
        # Postfix queue
        Postfix:
          checks:
            - type: queue_size
              warn: 5
              crit: 10
        '''



import os

from sauna.plugins import Plugin
from sauna.plugins import PluginRegister

my_plugin = PluginRegister('Load')


@my_plugin.plugin()
class Load(Plugin):

    def __init__(self, config):
        super().__init__(config)
        self._load = None

    @my_plugin.check()
    def load1(self, check_config):
        return (self._value_to_status_less(self.load[0], check_config),
                'Load 1: {}'.format(self.load[0]))

    @my_plugin.check()
    def load5(self, check_config):
        return (self._value_to_status_less(self.load[1], check_config),
                'Load 5: {}'.format(self.load[1]))

    @my_plugin.check()
    def load15(self, check_config):
        return (self._value_to_status_less(self.load[2], check_config),
                'Load 15: {}'.format(self.load[2]))

    @property
    def load(self):
        if not self._load:
            self._load = os.getloadavg()
        return self._load

    @staticmethod
    def config_sample():
        return '''
        # Load average
        Load:
          checks:
            - type: load1
              warn: 2
              crit: 4
            - type: load5
              warn: 2
              crit: 4
            - type: load15
              warn: 2
              crit: 4
        '''



from sauna.plugins import Plugin, human_to_bytes, bytes_to_human,\
    PluginRegister

my_plugin = PluginRegister('Redis')


@my_plugin.plugin()
class Redis(Plugin):

    def __init__(self, config):
        super().__init__(config)
        try:
            import redis
            self.redis = redis
        except ImportError:
            from .. import DependencyError
            raise DependencyError(self.__class__.__name__, 'redis-py',
                                  'redis', 'python3-redis')
        self._redis_info = None

    @my_plugin.check()
    def used_memory(self, check_config):
        status = self._value_to_status_less(
            self.redis_info['used_memory'], check_config, human_to_bytes
        )
        output = 'Used memory: {}'.format(self.redis_info['used_memory_human'])
        return status, output

    @my_plugin.check()
    def used_memory_rss(self, check_config):
        status = self._value_to_status_less(
            self.redis_info['used_memory_rss'], check_config, human_to_bytes
        )
        output = 'Used memory RSS: {}'.format(
            bytes_to_human(self.redis_info['used_memory_rss'])
        )
        return status, output

    @property
    def redis_info(self):
        if not self._redis_info:
            host = self.config.get('host', 'localhost')
            port = self.config.get('port', 6379)
            r = self.redis.StrictRedis(host=host, port=port)
            self._redis_info = r.info()
        return self._redis_info

    @staticmethod
    def config_sample():
        return '''
        # Redis
        Redis:
          checks:
            - type: used_memory
              warn: 128M
              crit: 1024M
            - type: used_memory_rss
              warn: 128M
              crit: 1024M
          config:
            host: localhost
            port: 6379
        '''
